;;;; http-client.scm - Client API - felix
;
; Copyright (c) 2000-2004, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


(declare
  (fixnum)
  ;(no-bound-checks)
  (export 
   http:send-request http:GET http:POST) 
  (foreign-declare #<<EOF
#ifdef _WIN32
#include "WINSOCK2.H"
#else
#include <netdb.h>
#include <netinet/in.h>
#endif
#define NET_CONV(expr) ntohs((expr))
EOF
  ) )

(declare (uses extras srfi-1 regex tcp utils))

(require 'http-utils)
(require-for-syntax 'regex-case)


;; Client API:

(define url-rx
  (if (test-feature? 'pregexp)
      "([A-Za-z]+\\://)?([\\-_a-zA-Z0-9.]+)(\\:[0-9]+)?(/.*)?"
      "([A-Za-z]+\\://)?([-_a-zA-Z0-9.]+)(\\:[0-9]+)?(/.*)?") )

(define (parse-url-host-and-port url)
  (match (string-match url-rx url)
    [(_ serv host port path)
     (values
      host
      (cond [port (string->number (substring port 1 (string-length port)))]
            [serv (let ((port (getservbyname (substring serv 0 (- (string-length serv) 3)))))
                    (if (> port 0)
                        port
                        (error "invalid service" serv)))]
	    [else 80] )
      (or path "/") ) ]
    [else (values url 80 "/")] ) )

(define getservbyname 
  (foreign-lambda* int ([c-string serv])
    "struct servent *se = getservbyname(serv, \"tcp\");"
    "if(se == NULL) return(0);"
    "else return(NET_CONV(se->s_port));") )

(define (http:send-request req . more)
  (let-optionals more ([in #f]
		       [out #f] )
    (let* ([req (if (string? req) 
		    (http:make-request 'GET req '(("Connection" . "close")))
		    req) ] 
	   [as (remove (lambda (a) (string=? (car a) "content-length")) (http:request-attributes req))]
	   [url (http:request-url req)] 
	   [b (http:request-body req)] )
      (let*-values ([(host port path) (parse-url-host-and-port url)]
		    [(i o) (if (and in out)
			       (values in out)
			       (tcp-connect host port) ) ] )
	(fprintf o "~A ~A ~A\r\nHost: ~A:~A\r\n" 
		 (string-upcase (symbol->string (http:request-method req)))
		 path
		 (string-upcase (symbol->string (http:request-protocol req)))
		 host port)
	(for-each
	 (lambda (a)
	   (fprintf o "~A: ~A\r\n" (car a) (cdr a)) )
	 as)
	(if (string? b)
	    (begin
	      (fprintf o "Content-Length: ~A\r\n\r\n" (string-length b))
	      (display b o) )
	    (display "Content-Length: 0\r\n\r\n" o) )
	(let ([header (read-line i)])
	  (let ([a (http:read-request-attributes i)])
	    (values header a i o) ) ) ) ) ) )

(define (http:GET req)
  (let-values ([(header a i o) (http:send-request req)])
    (let ([data (read-all i)])
      (close-input-port i)
      (close-output-port o)
      data) ) )

(define (http:POST req #!optional (args '()))
  (let-values ([(header a i o) 
		(http:send-request 
		 (if (string? req)
		     (http:make-request 
		      'POST req '(("Connection" . "close") ("Content-type" . "text/plain"))
		      (string-intersperse args "\r\n") )
		     (begin
		       (http:request-body-set! req (string-intersperse args "\r\n"))
		       req) ) ) ] )
    (let ([data (read-all i)])
      (close-input-port i)
      (close-output-port o)
      data) ) )
