;
; Written by Ed Watkeys <edw@xmog.com>
;
; Copyright 2005 Transmogrify, LLC
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions
; are met:
;
; 1. Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;
; 2. Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the distribution.
;
; 3. Neither the name of the company nor the names of its contributors
; may be used to endorse or promote products derived from this
; software without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
; FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
; COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
; INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
; BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
; CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
; ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;

(declare (export
	  path->dict
	  dict->path
	  split-string
	  join-list
	  merge-lists
	  add-params-to-path))

(define (add-params-to-path path-string params)
  (string-append
   path-string
   "?"
   (join-list
    (map (lambda (x)
	   (string-append
	    (symbol->string (car x))
	    "="
	    (standard-out-filter (cdr x))))
	 params)
    "&"
    #f
    #f)))

(define (split-string str sep)
  (let loop ((s str) (new ""))
    (if (= (string-length s) 0)
	(if (= (string-length new) 0)
	    '()
	    (cons new '()))
	(let ((ch (string-ref s 0))
	      (rest (substring s 1 (string-length s))))
	  (if (char=? sep ch)
		 (if (= (string-length new) 0)
		     (loop rest "")
		     (cons new (loop rest "")))
		 (loop rest
		       (string-append new
				      (string ch))))))))

(define (join-list lis sep prefix suffix)
  (apply string-append
	 (let ((new-list
		(let loop ((l lis))
		  (if (null? l)
		      '()
		      (cons (car l)
			    (let ((rest (cdr l)))
			      (if (null? rest)
				  (if (string? suffix)
				      (cons suffix '())
				      '())
				  (cons sep
					(loop rest)))))))))
	   (if (string? prefix)
	       (cons prefix new-list)
	       new-list))))

(define (route-pattern r)
  (car r))

(define (route-dict r)
  (cadr r))

(define (elem-symbol? e)
  (or (symbol? e) (pair? e)))

(define (elem-symbol-key e)
  (if (pair? e)
      (car e)
      e))

(define (elem-symbol-in-filter e)
      (if (and (pair? e)
	       (not (null? (cdr e))))
	  (cadr e)
	  standard-in-filter))

(define (elem-symbol-out-filter e)
  (if (and (pair? e)
	   (not (null? (cdr e)))
	   (not (null? (cddr e))))
      (caddr e)
      standard-out-filter))

(define (standard-in-filter x)
  (or (string->number x)
      (string->symbol x)))

(define (standard-out-filter x)
  (cond ((symbol? x)
	 (symbol->string x))
	((number? x)
	 (number->string x))
	((string? x)
	 x)
	(else
	 "UNDEFINED")))

(define (add-to-dict k v path pat dict)
  (try-route path pat
	     (cons (cons k v)
		   dict)))

(define (match-string path pat dict)
  (if (string=? (car path) (car pat))
      (try-route (cdr path) (cdr pat) dict)
      #f))

(define (match-symbol path pat dict)
  (let* ((elem (car pat))
	 (key (elem-symbol-key elem))
	 (value ((elem-symbol-in-filter elem) (car path))))
    (if (null? value)
	#f
	(add-to-dict key value (cdr path) (cdr pat) dict))))

  (define (try-route path pat dict)
    (if (null? pat)
	(if (null? path)
	    dict
	    #f)
	(let ((elem (car pat)))
	  (cond
	   ((null? path)
	    #f)
	   ((string? elem)
	    (match-string path pat dict))
	   ((elem-symbol? elem)
	    (match-symbol path pat dict))
	   (else
	    #f)))))

(define (path->dict path routes)
  (if (null? routes)
      #f
      (let ((route (car routes)))
	(or (try-route path (route-pattern route) (route-dict route))
	    (path->dict path (cdr routes))))))

; (define (path-elem->number e)
;   (or (string->number e) '()))
  
; (define (number->path-elem e)
;  (or (and (number? e) (number->string e)) '()))

(define (dict->path dict routes)
  (if (null? routes)
      (values #f #f)
      (let* ((route (car routes))
	     (r-pat (route-pattern route))
	     (r-dict (route-dict route)))
	(call-with-values
	    (lambda () (match-pattern dict r-pat))
	  (lambda (path extras)
	    (if extras
		(let ((query-params (subtract-dicts extras r-dict)))
		  (if query-params
		      (values path query-params)
		      (dict->path dict (cdr routes))))
		(dict->path dict (cdr routes))))))))

(define (match-pattern dict pat)
  (match-pattern-loop dict pat '()))

(define (match-pattern-loop dict pat path)
  (if (null? pat)
	 (values (reverse path) dict)
      (let ((e (car pat)))
	(cond
	 ((string? e)
	  (match-pattern-loop dict (cdr pat) (cons e path)))
	 ((elem-symbol? e)
	  (let ((match (elem-matches-dict e dict)))
	    (if (not (null? match))
		(match-pattern-loop (remove-elem e dict) (cdr pat) (cons match path))
		(values #f #f))))))))
	  
(define (remove-elem e dict)
  (let ((key (elem-symbol-key e)))
    (let loop ((old dict) (new '()))
      (if (null? old)
	  new
	  (if (eq? (caar old) key)
	      (loop (cdr old) new)
	      (loop (cdr old) (cons (car old) new)))))))

(define (elem-matches-dict e dict)
  (if (string? e)
      e
      (let ((value (assq (elem-symbol-key e) dict)))
	(if value
	    ((elem-symbol-out-filter e) (cdr value))
	    '()))))

(define (subtract-dicts big small)
  (let next-key ((small small) (big big))
    (if (null? small)
	big
	(let find-key ((big big) (saved '()))
	  (if (null? big)
	      #f
	      (if (eq? (caar small) (caar big))
		  (if (equal? (cdar small) (cdar big))
		      (next-key (cdr small) (merge-lists (cdr big) saved))
		      #f)
		  (find-key (cdr big) (cons (car big) saved))))))))

(define (merge-lists a b)
  (if (null? a)
      b
      (merge-lists (cdr a) (cons (car a) b))))

; (define (url-decode s)
;   (let ((decoded-s (decode-list (string->list s) '())))
;     (if decoded-s
; 	(list->string (reverse decoded-s))
; 	#f)))

; (define (url-decode-list l new)
;   (define (hex-digit? ch)
;     (if (char-numeric? ch)
; 	#t
; 	(if (char-alphabetic? ch)
; 	    (if (char-ci<=? ch #\f)
; 		#t
; 		#f)
; 	    #f)))
;   (define (hex-digit-value ch)
;     (if (char-numeric? ch)
; 	(- (char->integer ch) (char->integer #\0))
; 	(+ (- (char->integer (char-downcase ch)) (char->integer #\a)) 10)))
;   (define (decode-digits l)
;     (if (null? l)
; 	#f ; End of string looking for digit 1.
; 	(let ((digit-1 (car l)))
; 	  (if (hex-digit? digit-1)
; 	      (if (null? (cdr l))
; 		  #f ; End of string looking for digit 2.
; 		  (let ((digit-2 (cadr l)))
; 			(if (hex-digit? digit-2)
; 			    (url-decode-list
; 			     (cddr l)
; 			     (cons (integer->char
; 				    (+ (* 16 (hex-digit-value digit-1))
; 				       (hex-digit-value digit-2)))
; 				   new))
; 			    #f))) ; Digit 2 is not hexadecimal.
; 	      #f)))) ; Digit 1 is not hexadecimal.
;   (if (null? l)
;       new
;       (let ((ch (car l)))
; 	(if (char=? ch #\%)
; 	    (decode-digits (cdr l))
; 	    (decode-list (cdr l) (cons (car l) new))))))

; (define (decode-query-string q)
;   (map (lambda (e) (map (lambda (e) (url-decode e))
; 			(split-string e #\=)))
;        (split-string q #\&)))
