;;;; http-server.scm - Server API - felix
;
; Copyright (c) 2000-2005, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


(declare
  (fixnum)
  ;(no-bound-checks)
  (export 
   http:make-server
   http:add-resource
   http:find-resource
   http:remove-resource
   http:content-parser 
   http:request-method-handler 
   http:fallback-handler
   http:write-response-header 
   http:write-error-response
   http:request-count-limit
   http:startup-hook
   http:error-hook
   http:log-hook
   http:parse-embedded-content
   http:current-request-count
   http:error-response-handler
   http:register-route
   http:register-controller
   http:invoke-controller
   http:link-to))

(declare (uses extras srfi-1 srfi-13 srfi-18 tcp))

(require-for-syntax 'regex-case)
(require 'http-utils)
(require 'http-controller)

;;; Constants:

(define-constant default-request-count-limit 10000)


;;; Helper routines:

(define debugging #f)

(define (dribble fstr . args)
  (critical-section
   (when debugging
     (fprintf (current-error-port) "[~?]~%~!" fstr args) ) ) )

(define (string->canonicalized-symbol str)
  (string->symbol (string-upcase str)) )

(define (protocol->string p)
  (if p
      (string-upcase (symbol->string p))
      "HTTP/1.1") )

(define exn-message (condition-property-accessor 'exn 'message))
(define exn-arguments (condition-property-accessor 'exn 'arguments))
(define exn-location (condition-property-accessor 'exn 'location))
(define http:request-count-limit (make-parameter default-request-count-limit))
(define current-number-of-threads 0)

(define http:error-hook
  (make-parameter 
   (lambda (ex)
     (##sys#with-print-length-limit
      256
      (lambda ()
	(fprintf (current-error-port)
		 "thread ~A terminated with exception: ~A: ~S~A~%"
		 (thread-name (current-thread))
		 (exn-message ex) (exn-arguments ex) 
		 (let ([loc (exn-location ex)])
		   (if loc (sprintf " (~S)" loc) "") ) ) ) ) ) ) )

(define (thread-fork thunk)
  (set! current-number-of-threads (add1 current-number-of-threads))
  (thread-start!
   (make-thread
    (lambda ()
      (handle-exceptions ex ((http:error-hook) ex) (thunk))
      (set! current-number-of-threads (sub1 current-number-of-threads)) ) ) ) )

(define current-server-protocol #f)
(define current-server-name #f)
(define current-server-location #f)
(define current-server-port #f)
(define request-count 0)

(define http:startup-hook (make-parameter (constantly #f)))

(define (http:current-request-count) request-count)


;;; Create server:

(define (http:make-server port #!key
			  [name "ChickenRulesOK"]
			  [location "localhost"] 
			  [protocol #f] 
			  [init (lambda _ #f)]
			  [accept #f] 
			  [backlog 40]
			  [method-hook #f] )
  (lambda dbg
    (let ([dbgmode (:optional dbg #f)])
      (fluid-let ([debugging (eq? dbgmode #t)]
		  [current-server-protocol protocol] 
		  [current-server-name name]
		  [current-server-location location] 
		  [current-server-port port] )
	(dribble "waiting for requests...")
	((http:startup-hook))
	(let ([listener (tcp-listen current-server-port backlog accept)])
	  (define (serve)
	    (let-values ([(in out)
			  (handle-exceptions exn
			      (begin (print-error-message exn (current-error-port))
				     (values #f #f) )
			    (tcp-accept listener)) ] )
	      (when in
		(thread-fork
		 (lambda ()
		   (let ([id (thread-name (current-thread))])
		     (let-values ([(_ you) (tcp-addresses in)])
		       (dribble "request ~A from ~A; ~A (of ~A) started..." 
				request-count you id current-number-of-threads)
		       (let loop ()
			 (let ([k (dispatch-request in out you)])
			   (set! request-count (add1 request-count))
			   (if k
			       (begin
				 (dribble "~A (of ~A) kept alive, reading next request..." 
					  id current-number-of-threads)
				 (loop) )
			       (dribble "~A finished." id) ) ) ) ) ) ) ) ) ) )
	  (init)
	  (let loop ([r 0] [t 0])
	    (if (< current-number-of-threads (http:request-count-limit))
		(serve)
		(thread-yield!) )
	    (loop r t) ) ) ) ) ) )

(define content-parsers (make-hash-table))

(define (dispatch-request in out addr)
  (handle-exceptions ex
      (begin
	(hard-close in out)
	(signal ex) )
    (let ([req (read-request in out)])
      (when req
	(handle-request req in out) )
      ((http:log-hook) req addr)
      (or (and req (keep-alive? req))
	  (begin
	    (close-connection in out) 
	    #f) ) ) ) )

(define http:log-hook (make-parameter noop))

(define (keep-alive? req)
  (let ([con (lookup-request-attribute "connection" (http:request-attributes req))]
	[http11 (eq? 'HTTP/1.1 (http:request-protocol req))] )
    (if con
	(or (string-ci=? con "keep-alive")
	    (and http11
		 (not (string-ci=? con "close")) ) )
	http11) ) )

(define (close-connection in out)
  (dribble "closing connection...")
  (close-output-port out)
  (close-input-port in) )

(define (hard-close in out)
  (tcp-abandon-port in)
  (tcp-abandon-port out)
  (close-input-port in)
  (close-output-port out) )

(define (read-line-perhaps in)
  (handle-exceptions _ #!eof
    (read-line in (http:read-line-limit)) ) )

(define (read-request in out)
  (let ([ln (read-line-perhaps in)])
    (if (eof-object? ln)
	(begin
	  (dribble "client closed connection.")
	  #f)
	(begin
	  (dribble "read: ~S" ln)
	  (match (and (not (eof-object? ln)) (string-split ln))
	    [(method url protocol)
	     (dribble "request: method=~A, url=~A, protocol=~A" method url protocol)
	     (let ([p (string->canonicalized-symbol protocol)]
		   [sp current-server-protocol] )
	       (cond [(and sp (not (eq? p sp)))
		      (http:write-unsupported-protocol-response out) 
		      #f]
		     [(http:read-request-attributes in) =>
		      (lambda (attrs)
			(dribble "attributes: ~S" attrs) 
			(http:make-request
			 (string->canonicalized-symbol method)
			 (regex-case url
			   ["http://(.+)" (_ url) url]
			   [else url] )
			 attrs
			 (or (and-let* ([ct (find-content-type attrs)])
			       ((find-content-parser (string->symbol ct))
				(let ([a (lookup-request-attribute "content-length" attrs)])
				  (and a (string->number a)) )
				attrs
				in) )
			     '() )
			 p) ) ]
		     [else (http:write-bad-request-response out) #f] ) ) ]
	    [_ (http:write-bad-request-response out) #f] ) ) ) ) )

(define (http:parse-embedded-content in)
  (let ([port (if (string? in) (open-input-string in) in)])
    (let ([attrs (http:read-request-attributes port)])
      (cons attrs
	    (or (and-let* (attrs
			   [ct (find-content-type attrs)])
		  ((find-content-parser (string->symbol ct))
		   (let ([a (lookup-request-attribute "content-length" attrs)])
		     (if a 
			 (fxmin (string->number a) (http:request-limit))
			 (http:request-limit)))
		   attrs
		   port) )
		(read-string #f port) ) ) ) ) )

(define (find-content-type attrs)
  (and-let* ([ct (lookup-request-attribute "content-type" attrs)])
    (let ([lst (string-split ct ";")])
      (and (pair? lst) (car lst)) ) ) )

(define (http:content-parser ct . p)
  (if (pair? p)
      (hash-table-set! content-parsers ct (car p))
      (hash-table-ref content-parsers ct) ) )

(define (find-content-parser type)
  (or (http:content-parser type)
      (lambda (size _ port) (read-string size port)) ) )

(define (lookup-request-attribute a attrs)
  (and-let* ([a (assoc a attrs)]) (cdr a)) )

(define (http:write-error-response code msg . port)
  (let ([m ((http:error-response-handler) code msg)])
    (fprintf (:optional port (current-output-port))
"~A ~A ~A\r
Server: ~A\r
Location: ~A\r
Cache-Control: no-cache\r
Content-Length: ~A\r
Content-Type: text/html\r
Connection: close\r
\r
~A\r
\r
"
      (protocol->string current-server-protocol) code msg
      current-server-name
      current-server-location
      (string-length m)
      m) ) )

(define http:error-response-handler
  (make-parameter
   (lambda (code msg)
     (sprintf "<h1>~A ~A</h1>" code msg) ) ) )

(define (http:write-bad-request-response . port)
  (http:write-error-response 400 "Bad Request" (:optional port (current-output-port))) )

(define (http:write-unsupported-protocol-response . port)
  (http:write-error-response 505 "HTTP Version Not Supported" (:optional port (current-output-port))) )

(define (http:write-not-found-response . port)
  (http:write-error-response 404 "Not Found" (:optional port (current-output-port))) )

(define (http:write-internal-error-response . port)
  (http:write-error-response 500 "Internal Server Error" (:optional port (current-output-port))) )

(define (http:write-response-header #!optional [code 200] [msg "OK"] [alist '()] 
				    [port (current-output-port)]
				    [protocol current-server-protocol] )
  (fprintf port
	   "~A ~A ~A\r
Server: ~A\r
"
	   (protocol->string protocol)
	   code msg
	   (default-property "Server" current-server-name alist) )
  (fprintf port "Location: ~A\r\n" (default-property "Location" current-server-location alist))
  (for-each
   (lambda (h)
     (let ([hd (car h)])
       (unless (or (string-ci=? "Server" hd) (string-ci=? "Location" hd))
	 (fprintf port "~A: ~A\r\n" hd (cdr h)))  ) )
   alist) )

(define (default-property name def props)
  (or (and-let* ([a (find (lambda (a) (string-ci=? name (car a))) props)])
	(cdr a) )
      def) )

(define (handle-request req in out)
  (let ([m (http:request-method req)])
    (dribble "handling ~A request..." m)
    (let ([h (http:request-method-handler m)])
      (if h
	  (parameterize ([current-input-port in]
			 [current-output-port out] )
	    (h req) )
	  (error "undefined request method handler" m (http:request-url req)) ) ) ) )

(define request-method-handlers '())

(define (http:request-method-handler m . h)
  (let ([a (assq m request-method-handlers)])
    (if (pair? h)
	(if a 
	    (set-cdr! a (car h))
	    (set! request-method-handlers (alist-cons m (car h) request-method-handlers)) )
	(and a (cdr a)) ) ) )

(define (get/post-handler req)
  (let-values ([(loc args) (http:decode-url (http:request-url req))])
    (dribble "loc: ~S" loc)
    (dribble "args: ~S" args)
    (handle-exceptions
     ex
     (http:write-internal-error-response)
     (if (not (http:invoke-controller loc args req))
	 (let ((h (http:find-resource loc)))
	   (dribble "no controller invoked")
	   (if h
	       (h req args)
	       ((http:fallback-handler) req)))))))

(define http:fallback-handler
  (make-parameter
   (lambda (r)
     (http:write-not-found-response) ) ) )

(http:request-method-handler 'GET get/post-handler)
(http:request-method-handler 'POST get/post-handler)


;;; Handling resources:

(define resource-table (make-hash-table string=?))

(define (http:add-resource url h)
  (let ([urls (map ->string (if (pair? url) url (list url)))])
    (for-each (cut hash-table-set! resource-table <> h) urls) ) )

(define (http:remove-resource url)
  (let ([urls (map ->string (if (pair? url) url (list url)))])
    (for-each (cut hash-table-remove! resource-table <>) urls) ) )

(define (http:find-resource url)
  (hash-table-ref resource-table url) )

;;; Builtin content parsers:

;(http:content-parser 'text/xml (lambda (_ _ port) (ssax:xml->sxml port '())))

(http:content-parser 
 'application/x-www-form-urlencoded
 (lambda (size _ port)
   (let ([data (string-split (read-string size port) "&")])
     (map (lambda (def)
	    (regex-case def
	      ["([^=]+)=(.*)" (_ name value)
	       (cons name (http:canonicalize-string value)) ] 
	      [else (cons def #f)] ) )
	  data) ) ) )

(define read-multipart/form-data
  (let ([rx (regexp "boundary=([^ ;\n\r\t]+)")])
    (lambda (size attrs port)
      (let ([raw (read-string size port)])
	;;(fprintf (current-error-port) "raw: ~S~%" raw)
	(match (string-search rx (alist-ref "content-type" attrs string-ci=?))
	  [(_ b)
	   (let ([brx (regexp (conc "--" b "(--)?\r\n"))])
	     (let loop ([cs '()] [i 0])
	       (match (string-search-positions brx raw i)
		 [((start end) _)
		  (loop (if (zero? i) cs (cons (substring raw i start) cs))
			end) ]
		 [_ (reverse (map http:parse-embedded-content cs))] ) ) ) ]
	  [_ raw] ) ) ) ) )

(http:content-parser 'multipart/form-data read-multipart/form-data)

;;; HTTP-CONTROLLER GLUE BELOW

(define route-list '())

(define (http:register-route pattern params)
  (set! route-list (cons (list pattern params) route-list)))

(define controller-table (make-hash-table))

(define (http:register-controller name proc)
  (hash-table-set! controller-table name proc))

(define (find-controller controller)
  (hash-table-ref controller-table controller))

(define (http:invoke-controller url params req)
  (let ((params-dict (path->dict (split-string url #\/) route-list)))
    (dribble "params-dict: ~S" params-dict)
    (if params-dict
	(let* ((params-dict (merge-lists params-dict params))
	       (controller (assoc 'controller params-dict)))
	  (if controller
	      (let ((controller (find-controller (cdr controller))))
		(if controller
		    (controller params-dict req)
		    #f))
		#f))
	#f)))

(define (http:link-to params)
  (let-values (((path params) (dict->path params route-list)))
    (if path
	(let ((path-string (join-list path "/" "/" #f)))
	  (if (pair? params)
	      (add-params-to-path path-string params)
	      path-string)))))
