;;;; http-utils.scm - Utility routines - felix
;
; Copyright (c) 2000-2005, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


(declare
  (fixnum)
  (export 
   http:decode-url http:canonicalize-string
   http:read-line-limit http:request-limit
   http:read-request-attributes
   http:make-request http:request? http:request-url http:request-protocol http:request-attributes http:request-body http:request-method
   http:request-url-set! http:request-protocol-set! http:request-attributes-set! http:request-body-set! http:request-method-set!) )

(declare (uses srfi-1 srfi-13 srfi-18 regex))

(require-extension regex-case)


;;; Constants:

(define-constant default-read-line-limit 1024)
(define-constant default-request-limit 1000000)


;;; I/O utilities:

(define http:read-line-limit (make-parameter default-read-line-limit))
(define http:request-limit (make-parameter default-request-limit))


;;; Requests:

(define-record http:request
  method				; symbol
  attributes				; ((string1 . string2) ...)
  url					; string
  protocol				; symbol
  body)					; string | #f

(define http:make-request
  (let ([make-http:request make-http:request])
    (lambda (method url . more)
      (let-optionals more ([attrs '()] [body '()] [protocol 'HTTP/1.0])
	(make-http:request method attrs url protocol body) ) ) ) )

(define http:read-request-attributes 
  (let ((rx (if (test-feature? 'pregexp) 
		"([\\-A-Za-z0-9]+):[ ]*([^ ].*)"
		"([-A-Za-z0-9]+):[ ]*([^ ].*)") ) )
    (lambda (port)
      (let ([limit (http:read-line-limit)]
	    (total (http:request-limit)))
	(let loop ([attrs '()] [prev #f] [cnt 0])
	  (let  ([ln (read-line port limit)])
	    (cond [(or (eof-object? ln) (string-null? ln)) attrs] ; signal error?
		  [(fx> cnt total) #f]
		  [(char-whitespace? (string-ref ln 0))
		   (set-cdr! prev (string-append (cdr prev) " " ln)) 
		   (loop attrs prev (fx+ cnt (string-length ln)))]
		  [else
		   (regex-case ln
		     [rx (_ a val)
			 (let ([a (cons (string-downcase a) val)])
			   (loop (cons a attrs) a (fx+ cnt (string-length ln))) ) ]
		     [else #f] ) ] ) ) ) ) ) ) )


;;; URL and string operations:

(define (http:decode-url url)
  (regex-case url
    ["([^?]+)\\?(.*)" (_ loc args)
     (values
      (http:canonicalize-string loc)
      (parse-encoded-arguments args) ) ] 
    [else (values (http:canonicalize-string url) '())] ) )

(define (http:canonicalize-string str)
  (let loop ([i 0] [str (string-translate str "+" " ")])
    (match (string-search-positions "%[0-9ABCDEFabcdef]{2}" str i)
      [((i1 i2))
       (loop 
	(fx+ i1 1)
	(string-append
	 (substring str 0 i1)
	 (string (integer->char (string->number (substring str (fx+ i1 1) i2) 16)))
	 (substring str i2 (string-length str)) ) ) ]
      [_ str] ) ) )

(define (parse-encoded-arguments args)
  (let ([vals (string-split args "&")])
    (map (lambda (def)
	   (regex-case def
	     ["([^=]+)=(.*)" (_ name val)
	      (cons (http:canonicalize-string name) (http:canonicalize-string val))]
	     [else (cons (http:canonicalize-string def) "")] ) )
	 vals) ) )
